//! \addtogroup fefrontend
//! \{
//!

#ifndef _DynamicDataBuffer_h
#define _DynamicDataBuffer_h

#include<vector>

using namespace std ;

template<class T> class DynamicDataBuffer ;

typedef DynamicDataBuffer<float> 	FeatureBuffer ;	//!< Dynamic Data Buffer for feature extraction propose

//! Template for a buffer based on standard library vector class. 

//! It can contain elements of different type (int, char, float, etc.) and it can be a buffer of vectors or just single valued buffer. Also it can be a linear buffer or circular buffer; if the buffer is circular, the size has to be power of 2 so that % operation can be substituted for faster & operation. Another singular advantatge is that allows to the external user work with a linear indexing variable when refering to the buffer elements.
//!
//! This class allows some kind of processing of the data stored in the buffer. In the current, the following possibilities are considered:
//! - filter the vectors of stored in the buffer with any filter
//! 
//! Some approaches have been designed for our particular audio processing propose (for FeatureBuffer)
//! - For SAD propose: compute a static + delta + acceleration vector of the requested index, if is static
//! - For AED propose: compute a non-perceptual features + delta + acceleration + perceptual features + SBLETD of a non-perceptual + perceptual features vector 
//! 
//! In the following the most important methods are described.
//!
template<class T>
class DynamicDataBuffer : private vector<T> { // assumes cirQSize is power of 2 for subst. of % for &
 
public:
  DynamicDataBuffer(int dimension = 1, long cirQueueSize = 0); //!< Constructor
  ~DynamicDataBuffer() ; 													//!< Destructor 

  bool isCirQ() 				//! Return true if buffer is circular. False otherwise.
  { return cirQSize > 0 ; }								
  long getCirQSize()			//! Get circular size (if is circular). Returns O if not.
  { return cirQSize ; }							
  void setupCirQueue(int dimen, long cirQS)		//! Setup vector dimension and circular size.
  {						
      vecDimen = dimen ;
      cirQSize = cirQS ;
      vector<T>::resize(cirQS * dimen) ;	//resize vector
  }

  T& value(long index) 					//! Returns index-th value
  { 
      if(isCirQ()) index &= (cirQSize-1) ;	//modulus indexation needed if is circular
      return *(vector<T>::begin() + index) ;	
  }

  T* vec(long index) 					//! Returns pointer to index-th vector
  { 
      if(isCirQ()) index &= (cirQSize-1) ;	//modulus indexation needed if is circular
      return &(*(vector<T>::begin() + index * vecDimen)) ;
  }
  
  void resize(long n) 					//! Sets the length of the buffer to the new value 
  {
      if(isCirQ()) cirQSize=n ;		//set cirQSize to new value
      n *= vecDimen ;
      vector<T>::resize(n) ;			//resize vector
  }

  void reset() { last = 0 ; }			//!< Resets buffer

  void resetVecDimen(int dimen) ;	//!< Sets the vector dimension of the buffer to the new value.

  int getVecDimen() { return vecDimen ; }		//!< Get vec dimension

  //! Appends n values to the buffer. Requires vecDimen == 1.
  void appendVals(T* vals, int n)  
  //! \param vals values to append
  //! \param n number of values to append
  {
		if((!isCirQ()) && ((unsigned)(last + n) >= vector<T>::size()))	//if is NOT circular, resize vector before append, if required
      	resize(last + n);
      for(int i=0 ; i<n ; i++) value(last+i) = vals[i] ;	//add all values 
      last += n ;		//increase last with n
  }

  //! Appends vector of default dimmension. Note that could also be used with vecDimen == 1.
  void appendVec(T* vec) 
  //! \param vec vector to append
  { 
      if((!isCirQ()) && ((unsigned long)(last * vecDimen) >= vector<T>::size())) //if is NOT circular, resize vector before append, if required
      	resize(last + 32) ;	//Caution: a power of 2 increment is needed (32 is choosen by the author)
      long pos = last ;	//pos where to start writing
      if(isCirQ()) pos &= (cirQSize-1) ; //modulus indexation needed if is circular
      T* v = &(*(vector<T>::begin() + pos * vecDimen)) ;	//get pointer from first position
      for(int i=0 ; i<vecDimen ; i++) v[i] = vec[i] ;	//add all values
      last++ ;		//increase last with 1
  }


  long length() { return (long)last ; } 	//!< Returns length of buffer values

  bool isIndexValid(long index) { return index < last ; } //!< Index smaller than last ? = true. False otherwise. 

  float getDataRate() { return dataRate ; }			//!< Return dataRate private value
  void setDataRate(float freq) { dataRate=freq ; }	//!< Set dataRate to the new value freq

  // FeatureBuffer
  void creaHtkDelAccFilter(short halfDeltaSpan = 2, short halfAccelSpan = 1) ; //!< Creates the delta and acceleration filters of the desired size.
	void  creaSCFilter(short halfDeltaSpan =2);
  void timeFirFilter(long index, float* statDelAccVec, float* filCoef, short halfFilLength) ;//!< Filters the vector values of the requested index 
	
	void SCFilter(long index, float* statDelAccVec, float* filCoef, short halfFilLength);
	void ShiftDeltaFilter(long index,float* filOut,int P, int D, int K);

  bool getStatDelAccVector(long index, float* statDelAccVec) ;//!< Calculates and provides static + delta + acceleration vector of the requested index
	bool getStatDelAccVectorFlush(long index, float* statDelAccVec) ; 
	bool getDelAccVector(long index, float* statDelAccVec) ;//!< Calcul
	bool getDelVectorFlush(long index, float* statDelAccVec) ;//!< Calcul
	bool getDelAccVectorFlush(long index, float* statDelAccVec);
  bool getStat2DelAccVector(long index, float* featStatisticsVec);//!< Calculates and provides non-perceptual features + delta + acceleration + perceptual + SBLETD from the non-perceptual+perceptual vector of the requested index. 
  bool getStat2DelAccFlush(long index, float* featStatisticsVec);

	bool getStatShiftDelVectorFlush(long index,float* filOut,int P,int D,int K);
	bool getStatShiftDelVector(long index,float* filOut,int P,int D,int K);

	bool getShiftDelVector(long index,float* filOut,int P,int D,int K);
	bool getShiftDelVectorFlush(long index,float* filOut,int P,int D,int K);

	bool getAllFeatures(long index, float* statVec);
	//bool getSomeFeatures(long index, float* inVec,int num);
	bool getSomeFeatures(long index, float* inVec,int initf,int endf);
	bool getSCFeatures(long index, float* inVal);

  void saveSelFFtoHTK(const char* fname, bool saveDelAc);
  
	void saveFeatures(const char* fname, bool saveDelAcc) ;//!< Saves all features, static only, or static + delta + acceleration.
	//void saveHTKFeatures(const char* fname, bool saveDelAcc,bool doFeatWarping,string stringtable);
	void saveHTKFeatures(const char* fname, bool saveDelAcc);
  void saveEnergy(const char* fname, bool saveDelAcc,bool doFeatWarping,string stringtable);
	void saveHTKShiftFeatures(const char* fname,bool saveShiftDel,bool doFeatWarping,string stringtable);
	void saveFeatures(const char* fname, bool saveDelAcc, int* selFeatIdx, int selFeatSize) ;//!< Saves selected features, static only, or static + delta + acceleration.
	void saveHTK_SDC_Features(const char* fname,bool saveShiftDel,int D,int K,int P);
  void subBandLogEnergyTimeDifferences(long index, float* vecOut) ;//!< //Provides the time differences between the sub-band log energy of the requested index vector and the previous one

 protected:



 private:
  long last ;			//!< Index of last element in vector
  long cirQSize ;		//!< Circular size
  int vecDimen ; 		//!< Vector dimension

  float dataRate ; 	//!< Data sample rate in Hz
  
  short halfDelSpan ; 	//!< Half the length of delta filter
  short halfAccSpan ; 	//!< Half the lenght of acceleration filter
  float* htkDelFilter ;	//!< Delta filter
  float* htkAccFilter ;	//!< Acceleration filter
} ;


//! \}
//!

//Constructor
//! To instantiate an object of type DynamicDataBuffer. Returns an object of type DynamicDataBuffer.
//! \param dimension buffer's element size (higher than 1 means vector buffer)
//! \param cirQueueSize size of circular buffer; if zero then linear buffer. Circluar buffer is usually used for online implementation.
template<class T>
DynamicDataBuffer<T>::DynamicDataBuffer(int dimension, long cirQueueSize)
{
    vecDimen = dimension ;
    cirQSize = cirQueueSize ;
    last = 0 ;
    htkDelFilter = htkAccFilter = NULL ;
    dataRate = 0 ;

    if(cirQueueSize>0)
		vector<T>::resize(cirQueueSize * dimension) ;
}

//Destructor 
//! Free all memory used by the object and destroy the object.
template<class T>
DynamicDataBuffer<T>::~DynamicDataBuffer()
{
    vector<T>::clear() ;
    if(htkDelFilter != NULL)
	delete [] htkDelFilter ;
    if(htkAccFilter != NULL)
	delete [] htkAccFilter ;
}


//Sets the vector dimension of the buffer to the new value.
//! \param dimen new vector dimension 
template<class T>
void DynamicDataBuffer<T>::resetVecDimen(int dimen)
{
    int oldVecDimen = vecDimen, numVec=0 ;
    vecDimen = dimen ;
    if(isCirQ()) {
	resize(cirQSize) ;
	if(last)  numVec = cirQSize ;
    }
    else {
	numVec = last ;
	resize(numVec) ;
    }

    if(dimen > oldVecDimen) {
	for(int i=(numVec - 1) ; i>=0 ; i--) {
	    T* oldV = &(*(vector<T>::begin() + i * oldVecDimen)) ;
	    T* v = vec(i) ;
	    for(int j=0 ; j<dimen ; j++)
		v[j] = (j<oldVecDimen ? oldV[j] : (T)0) ;
	}
    } else if(dimen < oldVecDimen) {
	for(int i=0 ; i<numVec ; i++) {
	    T* oldV = &(*(vector<T>::begin() + i * oldVecDimen)) ;
	    T* v = vec(i) ;
	    for(int j=0 ; j<dimen ; j++)
		v[j] = oldV[j] ;
	}
    }
}


#endif // _DynamicDataBuffer_h


