
#define _FeatExtractorFF_
//
// c++ libraries
#include<fstream>
#include<iostream>
//
// local libraries
#include "FeatExtractorFF.h"
#include "ParamTable.h"

#define USE_CUDA2
//#define USE_CUDA2
#include <cufft.h>
#include <cuda.h>
#include <cuda_runtime_api.h>

#  define CUDA_SAFE_CALL_NO_SYNC( call) do {                                 \
    cudaError err = call;                                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)

#  define CUDA_SAFE_CALL( call) do {                                         \
    CUDA_SAFE_CALL_NO_SYNC(call);                                            \
    cudaError err = cudaThreadSynchronize();                                 \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)

#  define CUFFT_SAFE_CALL( call) do {                                        \
    cufftResult err = call;                                                  \
    if( CUFFT_SUCCESS != err) {                                              \
        fprintf(stderr, "CUFFT error in file '%s' in line %i.\n",            \
                __FILE__, __LINE__);                                         \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)

//
using namespace std;
//
//
// ---------------- Class MEL FILTER BANK  ----------------- //
class MelFBank {
 public:
    MelFBank(unsigned short numCh=3, unsigned short fftL=512, int startFr=0, 
	     int sampFr=8000, bool normalize=false) ; // constructor
    ~MelFBank(); // destructor

    void applyFB(float *in);

 private:
    unsigned short fftLength, numChanns;
    int startFreq, sampFreq;
    bool norm;
    float *fbInput;
		
    struct fbWindow {
			unsigned short startBin;
			vector<float> *win;
    };	

    vector<fbWindow> fbWins; // initialized through the initializer list of melFBank constructor
};
// --------------- Constructor and Destructor "FeatExtractorFF" ---------------- //
//! Constructor
FeatExtractorFF::FeatExtractorFF(int sampFr, bool isOnln)
{
    sampFreq = sampFr ;
    isOnline = isOnln ;
		init() ;
}
//! Destructor
FeatExtractorFF::~FeatExtractorFF()
{
    delete [] circFrame ;
    delete [] wFrame ;
	
    if(doLogFbMeanRem) delete [] logFbMean ;
		
		
	if(doUtLogFbMeanRem) delete [] logFbMean ;
		if(doVNorm) delete [] logFbVar ;
    delete melFBextor ;
	delete melFBspec;
	delete [] coefFrame ;
	delete [] absFftFrame ;
    delete [] win ;
    free(defFnLogFbMean) ;
}
//
// ---------------- End Constructor and Destructor "FeatExtractorFF" ----------------- //
//! Set initial values to the variables of the object
//! depending on the user settings in the parameter file
void
FeatExtractorFF::init()
{
	int halfDelSpan, halfAccSpan ; // they are passed to feature buffer
	int featVecSize=0;
  defFnLogFbMean = (char*) malloc (12*sizeof(char)) ;
  strcpy(defFnLogFbMean, "./logFbMean") ;
	
  //initialization based on data from parameter table
	
	ParamTable* feParams = (globalParamTable())->searchScope("FeatExtractorFF") ;
  ParamEntry* pe ;
  if(feParams != NULL) {
		
		if((pe = feParams->search("frSize")) != NULL) frSize = pe->val.i ;//frame size ; has to be lower than inChunkReadSize
		else frSize = 240 ; // default for 16k
		if((pe = feParams->search("frShift")) != NULL) frShift = pe->val.i ;// frame shift
		else frShift = 80 ; // default for 8k
		if((pe = feParams->search("fftSize")) != NULL) fftSize = pe->val.i ;	// FFT size
		else fftSize = 512 ; // default for 16k
		if((pe = feParams->search("preemCoef")) != NULL) preemCoef = pe->val.f ;// preemphasis
		else preemCoef = 0.97 ;
		if((pe = feParams->search("fbSize")) != NULL) fbSize = pe->val.i ;// mel filter-bank size
		else fbSize = 12 ;
		if((pe = feParams->search("melStartFreq")) != NULL) melStartFreq = pe->val.i ;// mel filter-bank start frequency
		else melStartFreq = 0 ;
		if((pe = feParams->search("halfDelSpan")) != NULL) halfDelSpan = pe->val.i ;//delta and acc htk filter half lengths
		else halfDelSpan = 2 ;
		if((pe = feParams->search("halfAccSpan")) != NULL) halfAccSpan = pe->val.i ;
		else halfAccSpan = 1 ;
		if((pe = feParams->search("melNorm")) != NULL) melNorm = (bool) pe->val.i ;// normalize mel filter-bank windows ?
		else melNorm = false ;
		if((pe = feParams->search("fbFloor")) != NULL) fbFloor = exp(pe->val.f) ;// mel filter-bank floor
		else fbFloor = exp(-50.0) ;
		if((pe = feParams->search("powSpecFloor")) != NULL) powSpecFloor = exp(pe->val.f) ;	// power spectrum floor
		else powSpecFloor = exp(-50.0) ;
		if((pe = feParams->search("logEnFloor")) != NULL) logEnFloor = exp(pe->val.f) ;// log energy floor
		else logEnFloor = exp(-50.0) ;
		if((pe = feParams->search("fnLogFbMean")) != NULL) fnLogFbMean = pe->val.s ;	// among of sub-band log energies
		else fnLogFbMean = defFnLogFbMean ;		
		if((pe = feParams->search("calcLowFreqFeat")) != NULL) calcLowFreqFeat = (bool) pe->val.i ;
		else calcLowFreqFeat = false ;	
		if((pe = feParams->search("calcHiFreqFeat")) != NULL) calcHiFreqFeat = (bool) pe->val.i ;
		else calcHiFreqFeat = false ;
		if((pe = feParams->search("calcZcr")) != NULL) calcZcr = (bool) pe->val.i ;
		else calcZcr = false ;	
		if((pe = feParams->search("calcSpecCen")) != NULL) calcSpecCen = (bool) pe->val.i ;
		else calcSpecCen = false ;	
		if((pe = feParams->search("calcSpecBand")) != NULL) calcSpecBand = (bool) pe->val.i ;
		else calcSpecBand = false ;	
		if((pe = feParams->search("doLogFbMeanRem")) != NULL) calcSbLogEn = (bool) pe->val.i ;// Remove Mean of FB
		else calcSbLogEn = false ;
		if((pe = feParams->search("doUtLogFbMeanRem")) != NULL) doUtLogFbMeanRem = (bool) pe->val.i ;
		else doUtLogFbMeanRem = false ;
		if((pe = feParams->search("doVNorm")) != NULL) doVNorm = (bool) pe->val.i ;	// FF features (0) or MFCCC (1)
		else doVNorm = false ;
		if((pe = feParams->search("FF0_DCT1")) != NULL) FF0_DCT1 = (bool) pe->val.i ;	
		else FF0_DCT1 = 0 ;
		if((pe = feParams->search("justEn")) != NULL) justEn = (bool) pe->val.i ;	
		else justEn = 0 ;
		if((pe = feParams->search("alpha_FF")) != NULL) alpha =  pe->val.f ;// Choose parameter alpha in kind A of FF
		else alpha = 1.0 ;if (alpha > 1 || alpha < 0 ) alpha=1.0 ;
		if((pe = feParams->search("coeffSize")) != NULL) coeffSize = pe->val.i ;	// output feature size 
		else coeffSize = fbSize;
		if((pe = feParams->search("ceplifter")) != NULL) ceplifter = pe->val.i ;	//Cepliftering, 0= means no liftering >0 applies liftering of order ceplifter
		else ceplifter = 0;
		if((pe = feParams->search("zero_coeff")) != NULL) zero_coeff = (bool) pe->val.i ;
		else zero_coeff = 0 ;
		if((pe = feParams->search("kind_ff")) != NULL) kind_ff =  pe->val.i ;
		else kind_ff = 3 ;
		if((pe = feParams->search("DCoffset")) != NULL) DCoffset = (bool) pe->val.i ;
		else DCoffset = 1 ;
		if((pe = feParams->search("addEnergy")) != NULL) addEnergy = (bool) pe->val.i ;
		else addEnergy = 0 ;
		if((pe = feParams->search("powSqrt")) != NULL) powSqrt = (bool) pe->val.i ;
		else powSqrt = 0 ;
		if((pe = feParams->search("remLoHiChan")) != NULL) remLoHiChan = (bool) pe->val.i ;
		else remLoHiChan = 0 ;
		if((pe = feParams->search("remLoHiFFcoef")) != NULL) remLoHiFFcoef = (bool) pe->val.i ;
		else remLoHiFFcoef = 0 ;
		if((pe = feParams->search("addlogEnergy")) != NULL) addlogEnergy = (bool) pe->val.i ;
		else addlogEnergy = 0 ;

  }
  else { // set all params default as for 16kHz
		frSize       = 480 ; 
		frShift      = 160 ; 
		fftSize      = 512 ; 	
		preemCoef    = 0.97 ; 
		fbSize       = 12 ; 
		melStartFreq = 0 ; 
		halfDelSpan  = 2 ;
		halfAccSpan  = 1 ;
		melNorm      = false ; 
		fbFloor      = exp(-50.0) ; 
		powSpecFloor = exp(-50.0) ;     
		logEnFloor   = exp(-50.0) ; 
		fnLogFbMean  = defFnLogFbMean ;
		calcLowFreqFeat = false ; 
		calcHiFreqFeat  = false ; 
		calcZcr = false ;
		calcSpecCen = false ;
		calcSpecBand = false ;
		calcSbLogEn = false ;		
		doLogFbMeanRem  = false ; 
		numSubBands  = 4 ;
		alpha=1.0 ;
		kind_ff = 3 ;
		FF0_DCT1 = 0 ;
		powSqrt = 0 ;
		addEnergy = 0 ;
		coeffSize    = fbSize;
		ceplifter    = 0 ;
		zero_coeff   = 0 ;
		DCoffset     = 1 ; 

  }
	// ---------------------------------------------------------------------------- //
	// initialization based on the above parameter setting
	frCounter = 0;
	circFrame = new float [frSize]; //circular frame space
	wFrame = new float [fftSize]; //linear frame space
	absFftFrame= new float [fftSize];
	win = new float [frSize];  //hamming window
	for (int i=0 ; i<frSize ; i++) win[i] = 0.54 - 0.46*cos(2*PI*i/(frSize - 1)) ;
	melFBextor = new MelFBank (fbSize, fftSize, melStartFreq, sampFreq, melNorm);
	melFBspec = new MelFBank (20, fftSize, melStartFreq, sampFreq, melNorm);
	// ---------------------------------------------------------------------------- //
	// normalization variables 
	if(doUtLogFbMeanRem){// utterance level filter bank mean removal
		logFbMean = new float [fbSize] ;// memory allocation for logFB mean
		for(int i=0 ; i<fbSize ; i++) logFbMean[i] = 0 ;
		frCounter = 0 ;
	}
	if(doVNorm){// variance normalization
		logFbVar = new float [fbSize] ;
		 for(int i=0 ; i<fbSize ; i++) logFbVar[i] = 0 ;
		frCounter = 0 ;
	}
  if(doLogFbMeanRem) { // log filter bank mean removal
		logFbMean = new float [fbSize] ; // memory allocation for logFB mean
		loadLogFbMean() ; // load logFB mean values from an external file
	#ifdef ACCUM_LOG_FB_MEAN
		frCounter = 0 ;
		accumLogFbMean = new double [fbSize] ;
		for(int i=0 ; i<fbSize ; i++) accumLogFbMean[i] = 0 ;
	#endif
  }
	// ---------------------------------------------------------------------------- //
	
	featVecSize = coeffSize ; //this is for DCT of MFCC
	
	if(remLoHiFFcoef){
		featVecSize = coeffSize-2 ; //this is for DCT of MFCC
	}else{
		featVecSize = coeffSize ; //this is for DCT of MFCC
	
	}
	

	
	
	if (zero_coeff) {
		if(addEnergy||addlogEnergy){
			cout << "WARNING: Energy or logEnergy and Zero coefficient can not be used simultaneously. Only logEn will be included." << endl ;
			zero_coeff = 0 ;
		}
		if(!FF0_DCT1){
			cout << "ff parameter " << FF0_DCT1 << endl;
			cout << "WARNING: Zero coefficient can not be used with FF parameters. Replaced by logEnergy." << endl ;
			zero_coeff = 0 ;
		}
		EnIdx = featVecSize ;	featVecSize++ ;
	}
	if(addEnergy){
		EnIdx = featVecSize ; featVecSize++ ;
	}
	if(addlogEnergy){
		logEnIdx = featVecSize ; featVecSize++ ;
	}
	if(calcLowFreqFeat){
		lffIdx = featVecSize ; featVecSize++ ;
	}
	if(calcHiFreqFeat){
		hffIdx = featVecSize ; featVecSize++ ;
	}
	if(calcZcr){
		zcrIdx= featVecSize ; featVecSize++ ;
	}
	if(calcSpecCen){
		specCenIdx= featVecSize ; featVecSize++ ;
	}
	if(calcSpecBand){
		specBandIdx= featVecSize ; featVecSize++ ;
	}
	if(calcSbLogEn){
		subBLEnergies= new float [numSubBands];
		numBandCoefs= (int)20/numSubBands; //numSubBands must be a mutiple of fbSize!
		sbLogEnIdx = featVecSize; featVecSize+=4; 
	}
	
	coefFrame    = new float [featVecSize];
	if(isOnline) {    // fe output === circular queue buffer with hardwired size
		feOutput.setupCirQueue(featVecSize, 32) ;
	}
	else {            // fe output === open-ended queue buffer
		if(feOutput.getVecDimen() != featVecSize) feOutput.resetVecDimen(featVecSize) ;
		if(fbOutput.getVecDimen() != fbSize ) fbOutput.resetVecDimen(fbSize) ;
	}

	if (FF0_DCT1){
		paramKind = 6 ;  //MFCC parameterization as defined by HTK
		if (halfDelSpan>0) paramKind += 0000400 ;
		if (halfAccSpan>0) paramKind += 0001000 ;
		if (addEnergy) paramKind += 0000100 ;
		if (zero_coeff) paramKind += 0020000 ;
	}
  else paramKind = 9 ; //User parameterization asdefined by HTK
  feOutput.creaHtkDelAccFilter(halfDelSpan, halfAccSpan) ;
  feOutput.setDataRate((float) sampFreq / frShift) ;
  reset() ;
}
//
// --------------------------------------------------------- // 
//
void
FeatExtractorFF::reset()
{
	// read first part of circFrame
	isFirstFrame = true ;
	frCounter= 0;
	// reset circFrame read & write position
	cfReadPos = cfWritePos = 0 ;
	// reset DC offset removal and preemphasis
	prevDCInSamp = prevDCOutSamp = prevPreemInSamp = 0 ;
	// reset output feature buffer (last=0)
	feOutput.reset() ;
	fbOutput.reset();
}
//
// --------------------------------------------------------- // 
//! Performs feature extraction for all the available input frames
//!
void
FeatExtractorFF::run()
{
	frCounter++;
	//cout << "counter: " <<frCounter << endl;
	if(isFirstFrame) {
		for(int i=0 ; i<(frSize-frShift) ; i++) circFrame[i] = (*feInput)[i] ;
		// DC offset and preemphasis at cfWritePos
		circFrameDCOffPreem (frSize-frShift) ;
		cfReadPos      = 0 ;
		cfWritePos     = frSize - frShift ;
		feInputReadPos = frSize - frShift ;
		feInputCanRead = feInput->size() - feInputReadPos;
		isFirstFrame = false ;
	}

	if(feInputCanRead < frShift) { // finish reading into circFrame from beginning of feInput
		unsigned short needRead = frShift - feInputCanRead ;
		for(int i=0 ; i<needRead ; i++)
			circFrame[(cfWritePos + i) % frSize] = (*feInput)[feInputReadPos + i] ;
		// DC offset and preemphasis starting at cfWritePos in circFrame
		circFrameDCOffPreem (needRead) ;
		cfWritePos     = (cfWritePos + needRead) % frSize ;
		feInputReadPos = needRead ;
		feInputCanRead = feInput->size() - feInputReadPos;
	
		// move from circFrame to wFrame with windowing and zero padding
		// also updates cfReadPos
		circFrame2wFrameWithWin() ;
		// process wFrame here
		procWinFrame() ;
	}

	while(feInputCanRead >= frShift) {
		for(int i=0 ; i<frShift ; i++)
			circFrame[(cfWritePos + i) % frSize] = (*feInput)[feInputReadPos + i] ;
		// DC offset and preemphasis at cfWritePos
		circFrameDCOffPreem (frShift) ;
		cfWritePos = (cfWritePos + frShift) % frSize ;
		feInputReadPos += frShift ;
		feInputCanRead = feInput->size() - feInputReadPos ;
	
		// move from circFrame to wFrame with windowing and zero padding
		// also updates cfReadPos
		circFrame2wFrameWithWin() ;	
		// process wFrame here
		procWinFrame() ;
	}

	for(unsigned int i=0 ; i<feInputCanRead ; i++)
		circFrame[(cfWritePos + i) % frSize] = (*feInput)[feInputReadPos + i] ;
	// DC offset and preemphasis at cfWritePos
	circFrameDCOffPreem (feInputCanRead) ;
	cfWritePos = (cfWritePos + feInputCanRead) % frSize ;
	feInputReadPos = 0 ;
	
}
//
// --------------------------------------------------------- // 
//! Move from circFrame to wFrame with windowing and zero padding
//!
void
FeatExtractorFF::circFrame2wFrameWithWin()
{
	// move from circFrame[frSize] to wFrame[fftSize] with windowing
	int j=cfReadPos ;
	for(int i=0 ; i<frSize ; i++) {
		wFrame[i] = win[i] * circFrame[j] ;
		if((++j) == frSize) j=0 ;
	}
	cfReadPos = (cfReadPos + frShift) % frSize ; // =j ???
	// zero padding
	for(int i=frSize ; i<fftSize ; i++) wFrame[i] = 0.0 ;
}
//
// --------------------------------------------------------- // 





















//! Process window frame
//!
void
FeatExtractorFF::procWinFrame()
{
	
// 	cout << "TSIGNAL" << endl;
// 	for(int i=0;i<frSize;i++){
// 		cout << i+1 << ":" << wFrame[i] << " " ;
// 	}
// 	cout << " " << endl;
	
	/*
	//time domain
	if(calcZcr) zeroCrossingRate() ;
	if(addEnergy) shortTimeEnergy();
	if(addlogEnergy) logEnergy() ;*/
	
	//copy data here
#ifdef USE_CUDA2

	int memsize = sizeof(float)*fftSize;
	CUDA_SAFE_CALL(cudaMalloc((void**)&cu_wFrame,memsize));
	CUDA_SAFE_CALL(cudaMemcpy(cu_wFrame,wFrame,memsize,cudaMemcpyHostToDevice));

#endif
// 	//FFT+powerSpechd
	realft(cu_wFrame,wFrame,fftSize, 1); // complex FFT in wFrame[0..fftSize-1]

#ifdef USE_CUDA2
	CUDA_SAFE_CALL(cudaFree(cu_wFrame));
#endif

	//Mel-scale filtering
	melFBextor->applyFB(wFrame); // FB output in wFrame[0..fbSize-1]
	
	//comento aixo per veure si va millor!!
	for (short i=0 ; i<fbSize ; i++) { // logarithm with flooring
		//comento aixo per veure si va millor	
		if(wFrame[i]<fbFloor) wFrame[i] = fbFloor ;
		wFrame[i] = log (wFrame[i]) ;
 	
	}
	
	
	
	melFBspec->applyFB(absFftFrame);
	for (short i=0 ; i<20 ; i++) { // logarithm with flooring
		if(wFrame[i]<fbFloor) absFftFrame[i] = fbFloor ;
	  absFftFrame[i] = log (absFftFrame[i]) ;
	}
	
	if(calcSbLogEn) subBandLogEnergies();
	if(doLogFbMeanRem) logFbMeanRemoval() ; // logFB mean removal
	
	if (!FF0_DCT1) {//Frequency filtering
		switch (kind_ff) { 
			case 0: 
				for (short j=0; j<coeffSize; j++) {coefFrame[j] = wFrame[j];}
				break ;
			case 1: 
				freqFilter('A') ;
				break ;
			case 2:
				freqFilter('B') ;
				break ;
			case 3:
				freqFilter('C') ; // frequency filter
				break ;
		}
		
		if(remLoHiFFcoef){
			for (short j=1; j<coeffSize-1; j++) {coefFrame[j-1] = wFrame[j];}
		}else{
			for (short j=0; j<coeffSize; j++) {coefFrame[j] = wFrame[j];}
		}
	} else {//DCT
		TrnCos(0,fbSize,coeffSize,wFrame,coefFrame) ; // DCT filter
		//liftering
		if (ceplifter>0)
		{
			
			for (short j=0; j<coeffSize;j++) {
				//coefFrame[j] *= (1+(ceplifter/2)*sin(M_PI*(j)/ceplifter));
				coefFrame[j] *= (1+(ceplifter/2)*sin(M_PI*(j+1)/ceplifter));
			}
		}
	}
	
	
	if(zero_coeff) coefFrame[EnIdx] = coefZERO ;
	if(addEnergy) coefFrame[EnIdx] = shortTimeEn;       // append energy
	if(addlogEnergy) coefFrame[logEnIdx] = logEn ; 	// append logenergy
	if(calcLowFreqFeat) coefFrame[lffIdx] = lowFreqFeat ; // append low-freq feature	
	if(calcHiFreqFeat) coefFrame[hffIdx] = hiFreqFeat ;  // append hi-freq feature
	if(calcZcr) coefFrame[zcrIdx] = zeroCrossRate ;
	if(calcSpecCen) coefFrame[specCenIdx]= specCentroid ;
	if(calcSpecBand) coefFrame[specBandIdx]= specBandwidth ;
	if(calcSbLogEn){
		for(int i=0;i<numBandCoefs;i++)
			coefFrame[sbLogEnIdx+i]=subBLEnergies[i] ;
	}
	/*cout << "trama" << endl;
	for(int i=0; i<coeffSize;i++){	
		cout << coefFrame[i] << endl;
	}*/
	feOutput.appendVec(coefFrame) ;//insereix nomes els que s'han determinat per el tamany o que? sino abans a AED amb wframe
//estava malament pq era del tamany de fftszie!

}


//
// --------------------------------------------------------- // 
// Feature extraction for one window frame
//
// Features for AED
// 
//		zeroCrossingRate() ;
// 		shortTimeEnergy();
// 		realft(wFrame, fftSize, 1); // complex FFT in wFrame[0..fftSize-1]
// 		powerSpec(); // power spectrum in wFrame[0..fftSize/2]
// 		spectralCentroid();
// 		spectralBandwidth();	
// 		//(calcLowFreqFeat || calcHiFreqFeat) lowHighFreqFeat() ; // calculate low- and hi-freq feature
// 		// second order power
// 		for(int i=0;i<=fftSize/2;i++) 
// 		{
// 			wFrame[i]=sqrt(wFrame[i]);
// 			absFftFrame[i]=wFrame[i];
// 		}
// 
// 		melFBextor->applyFB(wFrame); // FB output in wFrame[0..fbSize-1]
// 		for (short i=0 ; i<fbSize ; i++) { // logarithm with flooring
// 			if(wFrame[i]<fbFloor) wFrame[i] = fbFloor ;
// 			wFrame[i] = log (wFrame[i]) ;
// 		}
// 
// 		melFBspec->applyFB(absFftFrame);
//    		for (short i=0 ; i<20 ; i++) { // logarithm with flooring
// 	        	if(wFrame[i]<fbFloor) absFftFrame[i] = fbFloor ;
// 		        absFftFrame[i] = log (absFftFrame[i]) ;
// 		}
// 
// 		subBandLogEnergies();
// 		if(doLogFbMeanRem) logFbMeanRemoval() ; // logFB mean removal
// 		freqFilter('C') ; // frequency filter
// 		wFrame[logEnIdx] = shortTimeEn ;
// 		wFrame[zcrIdx]= zeroCrossRate ;
// 		wFrame[specCenIdx]= specCentroid ;
// 		wFrame[specBandIdx]= specBandwidth ;
// 		for(int i=0;i<numBandCoefs;i++)
// 			wFrame[sbLogEnIdx+i]=subBLEnergies[i] ;
// 		feOutput.appendVec(wFrame) ;
//
// Features for SPID
//
// 		logEnergy() ; 	// calculate frame energy into logEn from wFrame[]
// 		shortTimeEnergy();
// 		realft(wFrame, fftSize, 1); // complex FFT in wFrame[0..fftSize-1]
// 		powerSpec(); // power spectrum in wFrame[0..fftSize/2]
// 
// 		if(calcLowFreqFeat || calcHiFreqFeat) lowHighFreqFeat() ; // calculate low- and hi-freq feature
// 	
// 		melFBextor->applyFB(wFrame); // FB output in wFrame[0..fbSize-1]
// 	
// 		for (short i=0 ; i<fbSize ; i++) { // logarithm with flooring
// 			if(wFrame[i]<fbFloor) wFrame[i] = fbFloor ;
// 			wFrame[i] = log (wFrame[i]) ;
// 		}
// 		
// 		if(doLogFbMeanRem) logFbMeanRemoval() ; // logFB mean removal
// 		if (!FF0_DCT1) {
// 			switch (kind_ff) { 
// 			case 0: 
// 				for (short j=0; j<coeffSize; j++) {coefFrame[j] = wFrame[j];}
// 				break ;
// 			case 1: 
// 				freqFilter('A') ;
// 				break ;
// 			case 2:
// 				freqFilter('B') ;
// 				break ;
// 			case 3:
// 				freqFilter('C') ; // frequency filter
// 				break ;
// 			}
// 		 	for (short j=0; j<coeffSize; j++) {coefFrame[j] = wFrame[j];}
// 		} else 
// 		{ 
// 			TrnCos(0,fbSize,coeffSize,wFrame,coefFrame) ; // DCT filter
// 			//liftering
// 			if (ceplifter>0)
// 			{
// 				for (short j=0; j<coeffSize;j++) {
// 					coefFrame[j] *= (1+(ceplifter/2)*sin(M_PI*(j+1)/ceplifter));
// 				}
// 			}
// 		}
// 
// 
// 		if (zero_coeff) coefFrame[EnIdx] = coefZERO ;
// 		if (addEnergy) coefFrame[EnIdx] = shortTimeEn;       // append Energy
// 		if (addlogEnergy) coefFrame[logEnIdx] = logEn ; 	// append logEnergy
// 		if (calcLowFreqFeat) coefFrame[lffIdx] = lowFreqFeat ; // append low-freq feature
// 		if (calcHiFreqFeat)  coefFrame[hffIdx] = hiFreqFeat ;  // append hi-freq feature
// 		feOutput.appendVec(coefFrame) ;
//
// 
// 





//This is to process all the file stored at the input
//to be able to compute the mean at the frame level







//
// --------------- Feature Extraction Functions --------------- //
// 
// ZERO CROSSINGS
// Number of zero crossings of the waveform within a frame
void 
FeatExtractorFF::zeroCrossingRate()
{	
	zeroCrossRate=0;
	int currentSign;
	int shift;
	int i;
	for(i=1, shift=1;i<frSize;i++){
		currentSign=sign((float) wFrame[i]);	
		if( (currentSign==sign( (float) wFrame[i-shift]) )||(currentSign==0))
			shift++;
		else{ 
			zeroCrossRate++;
			shift=1;
		}
	}
	zeroCrossRate=(float)zeroCrossRate/frSize;
}
//
// SHORT TIME ENERGY
// Frame-level energy
void
FeatExtractorFF::shortTimeEnergy()
{
    shortTimeEn = 0 ;
    for(int i=0 ; i<frSize ; i++) shortTimeEn += (wFrame[i] * wFrame[i]) ;
    shortTimeEn=(float) shortTimeEn/frSize;
}
//
// LOG ENERGY
// Logarithmic version of the short-time energy feature
void
FeatExtractorFF::logEnergy()
{
    logEn = 0 ;
    for(int i=0 ; i<frSize ; i++) logEn += (wFrame[i] * wFrame[i]) ;
    if(logEn<logEnFloor) logEn = logEnFloor ;
		//cout << "logEnFloor" << logEnFloor << endl;
    logEn = log(logEn) ;
}

//
// FOURIER TRANSFORM
//
// Implemented according to "Numerical recipes in C++"
// - data[0..2*nn-1] array of nn complex numbers stored as 2*nn real numbers
// - replaces data[0..2*nn-1] by its FFT if isign==1, or by its IFFT if isign==-1
// - NOTE: the normalization factor 1/N not applied
void 
FeatExtractorFF::four1(float *data, short nn, short isign)
{
    unsigned long n, mmax, m, j, istep, i ;
    float wtemp, wr, wpr, wpi, wi, theta, tempr, tempi;

    n = nn << 1 ;
    j = 1 ;

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

    for (i=1 ; i<n ; i+=2) { // bit reversal section of the routine
        if (j > i) {
            SWAP(data[j-1],data[i-1]);
            SWAP(data[j],data[i]);
        }
        m = nn;
        while (m >= 2 && j > m) {
            j -= m;
            m >>= 1;
        }
        j += m;
    }
	//	printf("SWAP\n");
		//for (i=0 ; i<nn*2 ; i++) {
	//		cout << i <<":" << data[i] << " ";
	//	}
	//	cout << " " << endl;
		
    mmax = 2;
	//	printf("Limit %d",n);
    while (n > mmax) {
        istep = mmax << 1 ;
        theta = isign * ((float) TWOPI / mmax) ;
        wtemp = sin(0.5 * theta) ;
        wpr = -2.0 * wtemp * wtemp ;
        wpi = sin(theta) ;
        wr = 1.0;
        wi = 0.0;
				//cout << "wpr " << wpr << " wpi " << wpi << endl;
        for (m=1 ; m<mmax ; m+=2) {//doble de bucle; incr+2
            for (i=m ; i<=n ; i+=istep) {
                j = i + mmax;
                tempr = wr*data[j-1] - wi*data[j];
                tempi = wr*data[j] + wi*data[j-1];
							//	printf("xre %f + %f\n",data[j-1], data[j] );
                data[j-1] = data[i-1] - tempr;
                data[j] = data[i] - tempi;
							//	printf("dj %f + %f\n",data[j-1], data[j] );
                data[i-1] += tempr;
                data[i] += tempi;
							//	printf("dj %f + %f\n",data[i-1], data[i] );
            }
            wr=(wtemp=wr)*wpr-wi*wpi+wr;
            wi=wi*wpr+wtemp*wpi+wi;
        }
        mmax=istep ;
    }
}
#undef SWAP
//
// Implemented according to "Numerical recipes in C++"
// - data[0..n-1] array of n real numbers
// - replaces data[0..n-1] by positive freq. half of its FFT if isign==1, where
//   data[0] and data[1] contain the real first and last components of FFT
// - also calculates IFFT if isign==-1 assuming the format as above, but result
//   must be multiplied by 2/n
#ifdef USE_CUDA2


//N=N/2 length of a
//index<N

__global__ void  float2cufftComplex(float *a,cufftComplex *c,int N)
{
  int index = blockIdx.x*blockDim.x+threadIdx.x;
	if(index<N){
			c[index].x=a[2*index];
			c[index].y=a[2*index+1];
			//cout << h_data[i].x << " " << h_data[i].y  << endl;
	}

}

__global__ void  real2complex(float *a, cufftComplex *c, int N)
{
	unsigned int index = threadIdx.x + blockIdx.x*blockDim.x;
	if(index<N){
  	c[index].x = a[index];
  	c[index].y = 0.f;
	}
}


__global__ void  complex2real(cufftComplex *c, float *a, int N)
{

  int index = blockIdx.x*blockDim.x+threadIdx.x;
	if(index<N){
  	a[index]= c[index].x;
  }
}


//index
__global__ void  cu_powerSpec(cufftComplex *c, float *a,float powSpecFloor,int N)
{
  int index = blockIdx.x*blockDim.x+threadIdx.x;
	/*if(index<N){
		if(index==0){
  		a[2*index]= c[index].x;
			a[2*index+1]= c[N].x;
		}else{
			a[2*index]= c[index].x;
			a[2*index+1]= -c[index].y;
		}
  }*/
		//si c[index]the 0 es 0 de puta mare!!
	if(index<N){
		/*if(index==0){
  		a[2*index]= c[index].x;
			a[2*index+1]= c[N].x;
		}else{*/
			a[index]= c[index].x*c[index].x+c[index].y*c[index].y;
			if(a[index] < powSpecFloor) a[index] = powSpecFloor;

		//}
  }
		
	
}



void
FeatExtractorFF::realft(float *cu_data,float* backdata, short n, short isign)
{

	//create 1D plan
	cufftHandle plan;
	CUFFT_SAFE_CALL(cufftPlan1d(&plan,n, CUFFT_R2C, 1));
	//CUFFT_SAFE_CALL(cufftPlan1d(&plan,n, CUFFT_C2C, 1));
	//compute the execution configuration
	//cout << "memory" << sizeof(cufftComplex)*n << endl;
	//exit(0);
	//allocate memory for complex data
	cufftComplex *cu_complex;
  CUDA_SAFE_CALL(cudaMalloc ((void **) &cu_complex, sizeof(cufftComplex)*n));
	//transform from real to complex


	//real2complex<<<2,256>>>(cu_data,cu_complex,n);	
 	//CUFFT_SAFE_CALL(cufftExecC2C(plan,cu_data,cu_complex, CUFFT_FORWARD));
	

	CUFFT_SAFE_CALL(cufftExecR2C(plan,cu_data,cu_complex));
	//CUFFT_SAFE_CALL(cufftExecC2C(plan,cu_complex,cu_complex, CUFFT_FORWARD));
	//cufftcomplex2float<<<2,128>>>(cu_complex,cu_data,(n>>1));
	cu_powerSpec<<<2,256>>>(cu_complex,cu_data,powSpecFloor,n);
	//copy back to cpu
	CUDA_SAFE_CALL(cudaMemcpy(backdata,cu_data,sizeof(float)*n, cudaMemcpyDeviceToHost));

	CUFFT_SAFE_CALL(cufftDestroy(plan));
	CUDA_SAFE_CALL(cudaFree(cu_complex));

	//exit(0);
}//USE_CUDA
#endif





#ifdef USE_CUDA
void
FeatExtractorFF::realft(float *data, short n, short isign)
{
	cout << "realft cuda1" << endl;
	short i, i1, i2, i3, i4;
    float c1=0.5, c2, h1r, h1i, h2r, h2i, wr, wi, wpr, wpi, wtemp, theta;

	#ifdef USE_CUDA
		cufftHandle plan;
		cufftComplex *h_data, *d_data;
		cudaError_t err;
	
		int memsize = sizeof(cufftComplex)*(n>>1);
		h_data=(cufftComplex*)malloc(memsize);
		CUDA_SAFE_CALL(cudaMalloc((void**)&d_data,memsize));
		for (i=0 ; i< (n>>1)-1 ; i++) {
			(h_data[i]).x=data[2*i];
			(h_data[i]).y=data[2*i+1];
			//cout << h_data[i].x << " " << h_data[i].y  << endl;
		}
		//for (i=0 ; i< (n>>1)-1 ; i++) {
		//	cout << h_data[i].x << " " << h_data[i].y  << " ";
		//}
  	CUDA_SAFE_CALL(cudaMemcpy(d_data,h_data, memsize,cudaMemcpyHostToDevice));
		CUFFT_SAFE_CALL(cufftPlan1d(&plan,n>>1, CUFFT_C2C, 1));
	
	#endif

    theta = (float) PI / (float) (n>>1);

    if (isign == 1) {
        c2 = -0.5;

		#ifdef USE_CUDA
			CUFFT_SAFE_CALL(cufftExecC2C(plan,(cufftComplex *)d_data,(cufftComplex *)d_data,-1));
			CUDA_SAFE_CALL(cudaMemcpy(h_data,d_data,memsize,cudaMemcpyDeviceToHost));
		#else
			four1(data,(short) (n/2), 1);
		#endif
   } else {
        c2=0.5;
        theta = -theta;
    }
	
    wtemp = sin(0.5*theta);
    wpr = -2.0 * wtemp * wtemp;
    wpi = sin(theta);
    wr = 1.0 + wpr;
    wi = wpi;
	
 

	

    for (i=1 ; i<(n>>2) ; i++) {
		i2=1+(i1=i+i) ;
        i4=1+(i3=n-i1);	
		#ifdef USE_CUDA
			h1r = c1 * (h_data[i].x + h_data[(n>>1)-i].x);
			//h1r = -c1 * (h_data[i].x + h_data[(n>>1)-i].x);
			h1i = c1 * (h_data[i].y - h_data[(n>>1)-i].y);
			//h1i = -c1 * (h_data[i].y - h_data[(n>>1)-i].y);
			//h2r = -c2 * (h_data[i].y + h_data[(n>>1)-i].y);
			h2r = c2 * (h_data[i].y + h_data[(n>>1)-i].y);
			//h2i = c2 * (h_data[i].x - h_data[(n>>1)-i].x);
			h2i = -c2 * (h_data[i].x - h_data[(n>>1)-i].x);
			data[i1] = h1r - wr*h2r - wi*h2i;
			data[i2] = -h1i + wr*h2i - wi*h2r;
			data[i3] = h1r + wr*h2r + wi*h2i;
			data[i4] = h1i + wr*h2i + wi*h2r;

		#else
			h1r = c1 * (data[i1] + data[i3]);//0.5(F_n^e+F_{N/2-n}^e*)
			h1i = c1 * (data[i2] - data[i4]);//0.5(F_n^o-F_{N/2-n}^o*)
			h2r = -c2 * (data[i2] + data[i4]);//0.5(F_n^o-F_{N/2-n}^o*)
			h2i = c2 * (data[i1] - data[i3]);//-0.5(F_n^e-F_{N/2-n}^e*)
			data[i1] = h1r + wr*h2r - wi*h2i;
			data[i2] = h1i + wr*h2i + wi*h2r;
			data[i3] = h1r - wr*h2r + wi*h2i;
			data[i4] = -h1i + wr*h2i + wi*h2r;
		#endif	
        wr=(wtemp=wr)*wpr-wi*wpi+wr;
        wi=wi*wpr+wtemp*wpi+wi;
    }
    if (isign == 1) {
		#ifdef USE_CUDA
			data[0] = (h1r=h_data[0].x) + h_data[0].y;
			data[1] = h1r - h_data[0].y;
		#else
			data[0] = (h1r=data[0]) + data[1];
        	data[1] = h1r - data[1];
		#endif	
        
    } else {
        data[0] = c1 * ((h1r=data[0]) + data[1]);
        data[1] = c1 * (h1r - data[1]);
        four1(data, (short) (n/2), -1);
    }
	#ifdef USE_CUDA
		CUFFT_SAFE_CALL(cufftDestroy(plan));
		free(h_data);
		CUDA_SAFE_CALL(cudaFree(d_data));
	#endif
	/*cout << "data" << endl;
	for (i=0 ; i< n ; i++) {
			cout << i << ":" << data[i] << " " ; 
	}
	cout << " " << endl;*/
	
	//exit(0);
}//USE_CUDA
#endif

#ifdef USE_CUDA_v2



//------------------------------------------------------------------------//
__global__ void  real2complex(float *a, cufftComplex *c, int N)
{
	unsigned int index = threadIdx.x + blockIdx.x*blockDim.x;
	if(index<N){
  	c[index].x = a[index];
  	c[index].y = 0.f;
	}
}


__global__ void  complex2real(cufftComplex *c, float *a, int N)
{

  int index = blockIdx.x*blockDim.x+threadIdx.x;
	if(index<N){
  	a[index]= c[index].x;
  }
}

__global__ void  cufftcomplex2float(cufftComplex *c, float *a, int N)
{
  int index = blockIdx.x*blockDim.x+threadIdx.x;
	if(index<N){
		if(index==0){
  		a[2*index]= c[index].x;
			a[2*index+1]= c[N].x;
		}else{
			a[2*index]= c[index].x;
			a[2*index+1]= -c[index].y;
		}
  }
}
	


void
FeatExtractorFF::realft(float *data, short n, short isign)
{
	cout << "original data in host" << endl;

	//create 1D plan
	cufftHandle plan;
	CUFFT_SAFE_CALL(cufftPlan1d(&plan,n, CUFFT_C2C,1)); 
	
	//Compute the execution configuration


  //allocate device memory	
  float *rhs_real_d;
  cufftComplex *rhs_complex_d;
	//cufftComplex *tmp;

	cout << "memory" << sizeof(cufftComplex)*n << endl;
	exit(0);
  CUDA_SAFE_CALL(cudaMalloc ((void **) &rhs_complex_d, sizeof(cufftComplex)*n));
  CUDA_SAFE_CALL(cudaMalloc ((void **) &rhs_real_d   , sizeof(float)*n));
	//tmp=(cufftComplex*)malloc(sizeof(cufftComplex)*n);

  //copy audio to device
  CUDA_SAFE_CALL(cudaMemcpy(rhs_real_d,data,sizeof(float)*n, cudaMemcpyHostToDevice));
  //transform from real to complex	
  real2complex<<<4,128>>>(rhs_real_d,rhs_complex_d,n);

	CUDA_SAFE_CALL(cudaMemcpy(tmp,rhs_complex_d,sizeof(cufftComplex)*n, cudaMemcpyDeviceToHost));
	//cout << "complex data in device" << endl;
	//for (int i=0 ; i< n ; i++) {
	//		cout << i << ":" << tmp[i].x <<  "/" << tmp[i].y << " "  ; 
	//}	cout << " " << endl;

  CUFFT_SAFE_CALL(cufftExecC2C(plan,rhs_complex_d,rhs_complex_d, CUFFT_FORWARD));
	CUDA_SAFE_CALL(cudaMemcpy(tmp,rhs_complex_d,sizeof(cufftComplex)*n, cudaMemcpyDeviceToHost));
	//cout << "fft data in device" << endl;
	//for (int i=0 ; i< n ; i++) {
	//		cout << i << ":" << tmp[i].x <<  "/" << tmp[i].y << " "  ; 
	//}	cout << " " << endl;

	
	//cufftcomplex2float<<<2,256>>>(rhs_complex_d,rhs_real_d,n);
	cufftcomplex2float<<<2,128>>>(rhs_complex_d,rhs_real_d,(n>>1));
	CUDA_SAFE_CALL(cudaMemcpy(data,rhs_real_d,sizeof(float)*n, cudaMemcpyDeviceToHost));
/*cout << "notrick" << endl;
	for (int i=0 ; i< n ; i++) {
			cout << i << ":" << data[i] << " " ; 
	}
	cout << " " << endl;
	

	exit(0);*/
	CUFFT_SAFE_CALL(cufftDestroy(plan));
	CUDA_SAFE_CALL(cudaFree(rhs_real_d));
	CUDA_SAFE_CALL(cudaFree(rhs_complex_d));

}
#endif


//
// POWER SPECTRUM
// Computes the power spectrum from the real and imaginary values of the FFT
void 
FeatExtractorFF::powerSpec()
{
	int i, j, k ;
	
	wFrame[0] = wFrame[0] * wFrame[0];
	if (wFrame[0] < powSpecFloor) wFrame[0] = powSpecFloor;

	float tmp = wFrame[1] * wFrame[1];

	for(i=2, j=3, k=1 ; j<fftSize ; i+=2, j+=2, k++) {
		wFrame[k] = wFrame[i]*wFrame[i] + wFrame[j]*wFrame[j];
		if (powSqrt) wFrame[k] = sqrt(wFrame[k]) ;
		if (wFrame[k] < powSpecFloor) wFrame[k] = powSpecFloor;
	}
	wFrame[k] = tmp;
	if (powSqrt) wFrame[k] = sqrt(wFrame[k]) ;
	if (wFrame[k] < powSpecFloor) wFrame[k] = powSpecFloor;
}
//
// SPECTRAL CENTROID
// Frequency centroid of the (linear) spectrum. It is a measure of the spectral
// brightness
void 
FeatExtractorFF::spectralCentroid()
{	
	int freq;
	
	weightSpecSum=0;
	specSum=0;
	halfFftSize=fftSize/2;
	
	for(freq=0 ; freq<=halfFftSize;freq++){
		weightSpecSum+=(freq+1)*wFrame[freq];
		specSum+=wFrame[freq];
	}
	specCentroid=(float)weightSpecSum/specSum;
}
//
// SPECTRAL BANDWITH
// Frequency bandwith of the (linear) spectrum
void 
FeatExtractorFF::spectralBandwidth()
{	
	int freq;
	float sum;
 	float subs;	
	sum=0;
	
	halfFftSize=fftSize/2;
	for(freq=0 ; freq<=halfFftSize;freq++){
		subs=(freq+1)-specCentroid;
		sum+=subs*subs*wFrame[freq];
	}
	
	specBandwidth=sqrt(sum/specSum);
}
//
// SUB BAND LOG ENERGYES
// Computes the logEnergy of 4 sub-bans equally distributed along
// the Mel-Scaled filter bank.
// subBLEnergy[0...3] contents the four values, each for band.
void
FeatExtractorFF::subBandLogEnergies()
{

    int i, j, k;
    for(i=0, k=0; (i<numSubBands) && (k<20) ; i++,k+=j)
    {
          subBLEnergies[i]=0;
          for(j=0;j<numBandCoefs;j++)
                   subBLEnergies[i]+=absFftFrame[k+j];
    }
}
//
// FREQUENCY FILTERING 
void
FeatExtractorFF::freqFilter(char kind)
{
    int i ;
    float ffOut[2] ;
    
    switch (kind) {
	case 'A': // 1st order FF {1,-1} with low frequency energy preserved
	    for(i=1 ; i<fbSize ; i++)
		wFrame[i] = wFrame[i] - alpha * wFrame[i-1] ;
	    break ;
	case 'B': // 1st order FF {1,-1} with hi frequency energy preserved
	    for(i=0 ; i<(fbSize - 1) ; i++)
		wFrame[i] = wFrame[i+1] - wFrame[i] ;
	    break ;
	case 'C': // 2nd order FF {1,0,-1}
	    ffOut[0] = wFrame[1] ;
	    for(i=1 ; i<(fbSize - 1) ; i++) {
		ffOut[1] = wFrame[i+1] - wFrame[i-1] ;
		wFrame[i-1] = ffOut[0] ;
		ffOut[0] = ffOut[1] ;
	    }
	    ffOut[1] = -wFrame[i-1] ;
	    wFrame[i-1] = ffOut[0] ;
	    wFrame[i] = ffOut[1] ;
	    break ;
    }
}
//
int
FeatExtractorFF::sign(float sample)
{
//printf("s:%f ",sample);
int temp;
   temp= (sample == 0.0) ? 0 : ( (sample > 0.0) ? 1 : -1 );
//printf("r:%d ", temp);	
return temp;
}
//
void
FeatExtractorFF::lowHighFreqFeat()
{
    if(calcLowFreqFeat) {
		lowFreqFeat = wFrame[13] ;
		for(int i=14 ; i<=38 ; i++) lowFreqFeat += wFrame[i] ; // from 400 Hz to 1200 Hz for 16kHz samp rate
		lowFreqFeat /= 25.0 ;
		if(lowFreqFeat<logEnFloor) lowFreqFeat = logEnFloor ;
		lowFreqFeat = log(lowFreqFeat) ;
    }
    if(calcHiFreqFeat) {
		hiFreqFeat = wFrame[144] ;
		for(int i=145 ; i<=208 ; i++) hiFreqFeat += wFrame[i] ; // from 4500 Hz to 6500 Hz for 16kHz samp rate
		hiFreqFeat /= 65.0 ;
		if(hiFreqFeat<logEnFloor) hiFreqFeat = logEnFloor ;
		hiFreqFeat = log(hiFreqFeat) ;
    }
}
//
//
// ---------------- Class MEL FILTER BANK  ----------------- //
MelFBank::MelFBank(unsigned short numCh, unsigned short fftL, int startFr, 
		   int sampFr, bool normalize) : fbWins(numCh)
{
	unsigned i, j, k, halfConst;
	float normConst;
	vector<fbWindow>::iterator fbWindowIt;
	vector<float>::iterator floatIt;

	/* assign input params */
	fftLength=fftL ; numChanns=numCh ; startFreq=startFr ; sampFreq=sampFr ; norm=normalize ;

	vector<unsigned short> centrFreq (numChanns);
	
	/* Constants for calculation */
	float startFreq_mel    = 2595.0*log10(1.0 + (float)startFreq/700.0);
	float halfSampFreq_mel = 2595.0*log10(1.0 + (float)sampFreq/1400.0);

	for (i=0; i<numChanns; i++) {
		float freq = 700*(pow(10, (
		 + (float)i/((float)numChanns - 1)*(halfSampFreq_mel - startFreq_mel))/2595.0) - 1.0);
		centrFreq[i] = (unsigned short) (fftLength*freq/(float)sampFreq + 0.5);
	}

	fbWindowIt = fbWins.begin();

	/*-------------------*/
	/* first fb window 0 */
	/*-------------------*/
	//A half triangle window. First sample is the peak=1.
	fbWindowIt->startBin = centrFreq[0];
	fbWindowIt->win = new vector<float> (centrFreq[1] - centrFreq[0]);//length of the window (is half window actually)
	normConst = 0.0;
	halfConst = fbWindowIt->win->size();
	floatIt = fbWindowIt->win->begin();
	for(j=0 ; j<halfConst; j++, floatIt++) {
		(*floatIt) = (1.0 - (float)j/(float)halfConst);
		normConst += (*floatIt);
	}
	if (norm) {
		floatIt = fbWindowIt->win->begin();
		for (j=0; j<fbWindowIt->win->size(); j++, floatIt++) (*floatIt) /= normConst;
	}
    
	/*-----------------------------*/
	/* fb windows 1..(numChanns-2) */
	/*-----------------------------*/
  fbWindowIt++;
  for (i=1 ; i<(unsigned)(numChanns - 1) ; i++, fbWindowIt++) {
		fbWindowIt->startBin = centrFreq[i-1] + 1;
		fbWindowIt->win = new vector<float> (centrFreq[i+1] - centrFreq[i-1] - 1);
		normConst = 0.0;
		halfConst = centrFreq[i] - centrFreq[i-1] ;
		floatIt = fbWindowIt->win->begin();
		for (j=0 ; j<halfConst ; j++, floatIt++) {
				(*floatIt) = (float)(j + 1)/(float)halfConst ;
				normConst += (*floatIt);
		}
		halfConst = centrFreq [i+1] - centrFreq [i] ;
		for (j=(centrFreq[i] - centrFreq[i-1]), k=0 ; j<fbWindowIt->win->size() ; j++,k++, floatIt++) {
				(*floatIt) = 1.0 - (float)(k + 1)/(float)halfConst;
				normConst += (*floatIt) ;
		}
		if (norm) {
				floatIt = fbWindowIt->win->begin();
				for (j=0 ; j<fbWindowIt->win->size() ; j++) (*floatIt) /= normConst;
		}
  }
    
	/*------------------------------*/
	/* last fb window (numChanns-1) */
	/*------------------------------*/
	fbWindowIt->startBin = centrFreq[numChanns - 2] + 1;
	fbWindowIt->win = new vector<float> (centrFreq[numChanns - 1] - centrFreq[numChanns - 2]);
	normConst = 0.0;
	halfConst = fbWindowIt->win->size();
	floatIt = fbWindowIt->win->begin();
	for(j=0 ; j<halfConst; j++, floatIt++) {
		(*floatIt) = (float)(1 + j)/(float)halfConst;
		normConst += (*floatIt);
  }
  if (norm) {
		floatIt = fbWindowIt->win->begin();
		for (j=0; j<fbWindowIt->win->size(); j++, floatIt++) (*floatIt) /= normConst;
  }
    
}
//-------------------------------------------
MelFBank::~MelFBank()
{
    vector<fbWindow>::iterator fbWindowIt;

    fbWindowIt = fbWins.begin();
    for (unsigned short i=0 ; i<numChanns ; i++, fbWindowIt++)
	delete fbWindowIt->win ;
}
// -------------------------------------------

// 
void
MelFBank::applyFB(float *inSpec)
{

	vector<float> auxSum (numChanns);
	unsigned short i, j;
	vector<fbWindow>::iterator fbWindowIt;
	vector<float>::iterator floatIt;

	// Each filter bank channel is represented by a triangular window contained in fbWins
	fbWindowIt = fbWins.begin();//get first firter bank window
	//for(i=0 ; i<numChanns ; i++, fbWindowIt++) {//for all filter bank windows
	for(i=0 ; i<numChanns ; i++, fbWindowIt++) {//for all filter bank windows
		auxSum[i]=0.0;//init filter output
		floatIt = fbWindowIt->win->begin();//get first sample of window
		for(j=0 ; j<fbWindowIt->win->size() ; j++, floatIt++)//filter window
		//starting from startBin
	    auxSum[i] += inSpec[fbWindowIt->startBin + j] * (*floatIt);
  }

  for(i=0 ; i<numChanns ; i++) inSpec[i] = auxSum[i];
}



__global__ void  cu_applyFB(float *a,cufftComplex *c,int N)
{
  int index = blockIdx.x*blockDim.x+threadIdx.x;
	if(index<N){
			c[index].x=a[2*index];
			c[index].y=a[2*index+1];
			//cout << h_data[i].x << " " << h_data[i].y  << endl;
	}

}












// ------------------------------------------------------------------------------------- //
// DC offset removal and pre-emphasis on circFrame
void
FeatExtractorFF::circFrameDCOffPreem (int numSamples)
{
    int i, j=cfWritePos ;
    float curInSamp ;

    for(i=0 ; i<numSamples ; i++) {
		if (DCoffset)
		{
			// DC offset 
			// y[n] = x[n] - x[n-1] + 0.9990234375 * y[n-1] , note: 0.9990234375 = (1-1/1024)
			curInSamp = circFrame[j] ;
			circFrame[j] = curInSamp - prevDCInSamp + 0.9990234375 * prevDCOutSamp ;
			prevDCInSamp = curInSamp ;
			prevDCOutSamp = circFrame[j] ;
		}
		// Preemphasis
		// y[n] = x[n] - preemCoef * x[n-1]
		curInSamp = circFrame[j] ;
		circFrame[j] -= (preemCoef * prevPreemInSamp) ;
		prevPreemInSamp = curInSamp ;
	
		if((++j) == frSize) j=0 ;
    }
		//prevPreemInSamp = 0 ;
}
// logFB mean removal
void
FeatExtractorFF::utAcumLogFbMean()
{
  	frCounter++;
    for(int i=0 ; i<fbSize ; i++) {
		logFbMean[i] +=  wFrame[i] ; // accumulate for logFB mean during train
	}
}

void
FeatExtractorFF::utLogFbMeanCalc()
{
	//cout << "frCounter " << frCounter << endl;
    for(int i=0 ; i<fbSize ; i++) {
		logFbMean[i] = logFbMean[i] /frCounter ; // accumulate for logFB mean during train
		//cout << logFbMean[i] << "/" << frCounter << "=" << logFbMean[i] << endl;
	}
}

void
FeatExtractorFF::utLogFbVarCalc()
{
	double tmp;
	for(int j=0;j<fbSize;j++){
			logFbVar[j]=0.0;
		}
		
	if(frCounter!=fbOutput.length()){ cout << "Error in utLogFbVarCalc" << endl; exit(0);}	
	for(int i=0;i< fbOutput.length() ;i++){
		//logFbVar=fbOutput.vec(i);
		wFrame=fbOutput.vec(i);
		for(int j=0;j<fbSize;j++){
			tmp=wFrame[j]-logFbMean[j];
			logFbVar[j]+=tmp*tmp;
		}
	}
	for(int j=0;j<fbSize;j++){
			logFbVar[j]/=frCounter;
			logFbVar[j]=sqrt(logFbVar[j]);
			
			//cout << logFbVar[j] << " " ;
	}
	//cout << " " << endl;
}



void
FeatExtractorFF::utLogFbMeanRemoval()
{
		float diff, val;
    for(int i=0 ; i<fbSize ; i++){
		
		if(doVNorm){
		//diff=wFrame[i]-logFbMean[i];
		//val=diff/(diff*diff);
		wFrame[i] -= logFbMean[i];
		wFrame[i]/=logFbVar[i] ;;
		
		}else{
		
		wFrame[i] -= logFbMean[i] ;
		}
	}
}



void
FeatExtractorFF::logFbMeanRemoval ()
{
    float auxVal ;

#ifdef ACCUM_LOG_FB_MEAN
    frCounter++ ;
    for(int i=0 ; i<fbSize ; i++) {
	accumLogFbMean[i] += wFrame[i] ; // accumulate for logFB mean during train
#else
    for(int i=0 ; i<fbSize ; i++) {
#endif
	// remove mean
	auxVal = wFrame[i] - logFbMean[i] ;
	// update mean for next frame ; note: 0.9921875 = (1-1/128)
	logFbMean[i] = 0.9921875 * logFbMean[i] + 0.0078125 * wFrame[i] ;
	wFrame[i] = auxVal ;
    }
}

void
FeatExtractorFF::loadLogFbMean()
{
    int nBands, nRead ;
    FILE* f ;

    if((f = fopen(fnLogFbMean, "r")) == NULL) {
	printf("FeatExtractorFF::loadLogFbMean error opening file %s\n", fnLogFbMean) ;
	exit(0) ;
    }

    if((nRead=fscanf(f, "%i \t\n\014", &nBands)) && nRead!=EOF) {
	if(nBands != fbSize) {
	    printf("FeatExtractorFF::loadLogFbMean: error num of bands in logFbMean file (%i) differs from FB analysis order (%i)\n", nBands, fbSize) ;
	    exit(0) ;
	}
	float* pLogFbMean = logFbMean ;
	while((0 < nBands--) && (nRead=fscanf(f, "%f \t\n\014", pLogFbMean)) && (nRead != EOF))
	    pLogFbMean++ ;
	if((++nBands) != 0) {
	    printf("FeatExtractorFF::loadLogFbMean: error number of elements read (%i) differs from expected (%i)\n",fbSize-nBands, fbSize) ;
	    exit(0) ;
	}
    }
    else {printf("FeatExtractorFF::loadLogFbMean error reading num lines\n") ; exit(0) ;}
    fclose(f) ;
}
 
int FeatExtractorFF::getLowFreqFeatIdx() { return lffIdx ; }
int FeatExtractorFF::getHiFreqFeatIdx()  { return hffIdx ; }

#ifdef ACCUM_LOG_FB_MEAN
void
FeatExtractorFF::printAverLogFbMean()
{
    printf("Average logFB mean:\n") ;
    for(int i=0 ; i<fbSize ; i++)
	printf("%e ",(float) (accumLogFbMean[i] / frCounter)) ;
    printf("\n") ;
}
#endif

/***********************************************************************
 *	MelFreq: TrnCos
 *	Realiza la transformada discreta de coseno de una senhal real.
 *	
 *	SINTAXIS:
 *		int		TrnCos(
 *			int		Inv,
 *			int		Long,
 *			int		Orden,
 *			float	*Ent,
 *			float	*Sal);	
 *
 *	DESCRIPCION:
 *		Dada una senhal real de longitud Long, "TrnCos" calcula los
 *		Orden primeros coeficientes de su transformada ¿inversa? de
 *		coseno.
 *
 *	ENTRADA:
 *		Inv:
 *			Indicador de si se va a realizar la FFT Directa o Inversa.
 *			Los valores que puede tomar Inv son:
 *
 *			0: Para realizar la FFT Directa.
 *
 *			1: Para realizar la FFT Inversa.
 *
 *		Long:
 *			Número de puntos de la señal de entrada.
 *
 *		Orden:
 *			Número de puntos de la senhal de salida. Debe ser menor o
 *			igual a Long.
 *
 *		Ent:
 *			Senhal de entrada. Si Sal (la senhal de salida) apunta a la
 *			posicion de memoria cero, Ent almacenara al final la senhal
 *			transformada de salida.
 *
 *	SALIDAS:
 *		Sal:
 *			Transformada coseno de la senhal Sen. Si vale (float *) 0,
 *			se almacenara en Ent.
 *
 *	DIAGNOSTICOS:
 *		Si no se produce ningún error, "TrnCos" devuelve el entero 
 *		cero. En caso de error se devuelve un entero negativo.
 *
 *		Los errores que "TrnCos" es capaz de detectar son:
 *
 *		a.- Entrada Incorrecta: si "Long" es menor o igual a cero, si
 *		"Orden" es menor o igual a cero o mayor que "Long". Valor de
 *		Retorno: -1;
 *
 *		b.- Error de memoria: si no se puede ubicar la memoria
 *		necesaria para alojar la senhal temporal en el caso de que
 *		Sal apunte a cero. Valor de Retorno: -2;
 *
 *	NOTA:
 *		a.- La formula aplicada en en esta rutina ha sido extraida
 *		del libro "HTK: Hidden Markov Model Toolkit V1.5. Reference
 *		Manual", apartado 7.1.4 Implementation, pagina 73. En el momento
 *		de confeccionar la rutina, desconozco si esta en invertible y
 *		como. No obstante, y de cara a mantener en el futuro la
 *		compatibilidad p'atras, introduzco el parametro Inv, cuyo valor
 *		es ignorado actualmente.
 *
 *	DESARROLLO:
 *		Versión 1:
 *			Autor:	Albino Nogueiras Rodríguez
 *			Fecha:	6 de Noviembre de 1995
 **********************************************************************/


int	
FeatExtractorFF::TrnCos(
	int		Inv,			/* Indicador de FFT o FFT inversa		*/
	int		OrdAna,			/* Orden del analisis					*/
	int		NumCof,			/* Número de coeficientes de salida		*/
	float	*Ent,			/* señal de entrada						*/
	float	*Sal)			/* señal de salida						*/
{
	int		Cof, Ord;			/* Contadores						*/
	float	Factor;				/* PI / OrdAna						*/
	int		MMalloc = 0;	/* Indicador de utilización de "malloc"	*/

	/*
	 * Comprobamos que "OrdAna" y "NumCof" tengan un valor correcto.
	 */
	if (OrdAna <= 0 || NumCof <= 0) return -1;

	if (Sal == (float *) 0) {
		Sal = Ent;
		if ((Ent = (float *) malloc(OrdAna * sizeof(float))) == (float *) 0) {
			return -2;
		}
		if (memcpy((void *) Ent, (void *) Sal, OrdAna * sizeof(float)) == (void *) 0) {
			return -2;
		}
		MMalloc = 1;
	}

	switch (Inv) {
	case 0	:
		if(remLoHiChan) Factor = M_PI / (OrdAna-2);
		else Factor = M_PI / OrdAna;
		
		for (Cof = 0; Cof < NumCof; Cof++) {
			Sal[Cof] = 0.0;
			if(remLoHiChan){
				for (Ord = 1; Ord < OrdAna-1; Ord++) {
					//Sal[Cof] += Ent[Ord] * cos(Factor * (Cof + 1) * (Ord + 0.5));
					Sal[Cof] += Ent[Ord] * cos(Factor * (Cof + 1) * (Ord - 0.5));
				}
				Sal[Cof] *= sqrt(2.0 / (OrdAna-2));
			}else{
				for (Ord = 0; Ord < OrdAna; Ord++) {
					Sal[Cof] += Ent[Ord] * cos(Factor * (Cof + 1) * (Ord + 0.5));
				}
				Sal[Cof] *= sqrt(2.0 / OrdAna);
			}
			
			
		
			
		}
		break;
	case 1	:
		Factor = M_PI / (OrdAna - 1);
		for (Cof = 0; Cof < NumCof; Cof++) {
			Sal[Cof] = Ent[0];
			for (Ord = 1; Ord < OrdAna; Ord++) {
				Sal[Cof] += Ent[Ord] * cos(Cof * Ord * Factor);
			}
			Sal[Cof] *= sqrt(2.0 / (OrdAna - 1));
		}
		Sal[0] *= sqrt(0.5);
		break;
	case 2	:
		Factor = M_PI / OrdAna;
		for (Cof = 0; Cof < NumCof; Cof++) {
			Sal[Cof] = Ent[0] * sqrt(0.5);
			for (Ord = 1; Ord < OrdAna; Ord++) {
				Sal[Cof] += Ent[Ord] * cos((Cof + 0.5) * Ord * Factor);
			}
			Sal[Cof] *= sqrt(2.0 / OrdAna);
		}
		break;
	case 3	:
		Factor = M_PI / OrdAna;
		for (Cof = 0; Cof < NumCof; Cof++) {
			Sal[Cof] = Ent[0];
			for (Ord = 1; Ord < OrdAna; Ord++) {
				Sal[Cof] += Ent[Ord] * cos(Cof * (Ord + 0.5) * Factor);
			}
			Sal[Cof] *= sqrt(2.0 / OrdAna);
		}
		Sal[0] *= sqrt(0.5);
		break;
	case 4	:
		Factor = M_PI / OrdAna;
		for (Cof = 0; Cof < NumCof; Cof++) {
			Sal[Cof] = 0.0;
			for (Ord = 0; Ord < OrdAna; Ord++) {
				Sal[Cof] += Ent[Ord] * cos((Cof + 0.5) * (Ord + 0.5) * Factor);
			}
			Sal[Cof] *= sqrt(2.0 / OrdAna);
		}
		break;
	case 5	:
		Factor = M_PI / OrdAna;
		for (Cof = 0; Cof < NumCof; Cof++) {
			Sal[Cof] = 0.0;
			for (Ord = 0; Ord < OrdAna; Ord++) {
				Sal[Cof] += Ent[Ord] * sin((Cof + 1) * (Ord + 0.5) * Factor);
			}
			Sal[Cof] *= sqrt(2.0 / OrdAna);
		}
		break;
	default :
		fprintf(stderr, "Error en TrnCos: Inv debe valer entre 0 y 5\n");
		return -1;
		break;
	}

	if (MMalloc == 1) free((void *) Ent);
	
	//Calculo coefficiente 0

	coefZERO = 0.0;
	if(remLoHiChan){
		for (Ord = 1; Ord < OrdAna-1; Ord++) {
		//for (Ord = 1; Ord < OrdAna-1; Ord++) {
			coefZERO += Ent[Ord] ;
		}
		coefZERO *= sqrt(2.0 / (OrdAna-2));
	}else{
		for (Ord = 0; Ord < OrdAna; Ord++) {
		//for (Ord = 1; Ord < OrdAna-1; Ord++) {
			coefZERO += Ent[Ord] ;
		}
		coefZERO *= sqrt(2.0 / OrdAna);
	}

	return 0;

}


